import nd2
import numpy as np
from collections import OrderedDict

AXIS_ORDER = ('P', 'T', 'Z', 'C', 'S', 'Y', 'X')


def read(image):

    with nd2.ND2File(image) as image:
        raw_axis_sizes = image.sizes
        axis_order = tuple(axis for axis in AXIS_ORDER if axis in raw_axis_sizes.keys())
        axis_sizes = OrderedDict((axis, raw_axis_sizes[axis]) for axis in axis_order)

    return axis_sizes, axis_order


def parse(image, dask, axis_sizes, axis_order, overlap, slices):

    with nd2.ND2File(image) as image:

        if dask:
            image_array = image.to_dask()
        else:
            image_array = image.asarray()

        raw_axis_order = image.sizes.keys()
        image_array = np.moveaxis(image_array,
                                  range(len(raw_axis_order)), (axis_order.index(axis) for axis in raw_axis_order))

        if 'P' not in axis_order:

            tiles = np.empty((1, 1), dtype=object)
            tiles[0, 0] = image_array[slices]
            tiling = (1, 1)
            overlap = (0, 0)
            image_shape = image_array.shape

        else:

            positions = None

            for loop in image.experiment:
                if isinstance(loop, nd2.structures.XYPosLoop):
                    positions = np.array([point.stagePositionUm[:2] for point in loop.parameters.points]).T
                    break

            camera_transformation = np.array(image.metadata.channels[0].volume.cameraTransformationMatrix).reshape(2, 2)

            x, y = np.linalg.inv(camera_transformation) @ positions

            if positions[0, 1] - positions[0, 0] > 0:
                x = -x
            if positions[1, -1] - positions[1, 0] > 0:
                y = -y

            x_ndim = round(np.ptp(x) / abs(x[0] - x[1])) + 1
            y_ndim = round(np.ptp(y) / abs(x[0] - x[1])) + 1

            j = np.rint((x - min(x)) / (np.ptp(x) / (x_ndim - 1))).astype(int)
            i = np.rint((y - min(y)) / (np.ptp(y) / (y_ndim - 1))).astype(int)

            if overlap is None:
                x_overlaps = np.empty(0)
                y_overlaps = np.empty(0)
                for col in np.unique(j):
                    x_overlaps = np.append(x_overlaps, np.mean(x[np.where(j == col)]))
                for row in np.unique(i):
                    y_overlaps = np.append(y_overlaps, np.mean(y[np.where(i == row)]))
                x_overlap = round(1 - (np.mean(np.diff(x_overlaps)) / image.metadata.channels[0].volume.axesCalibration[
                    0]) / image.attributes.widthPx, 2)
                y_overlap = round(1 - (np.mean(np.diff(y_overlaps)) / image.metadata.channels[0].volume.axesCalibration[
                    1]) / image.attributes.heightPx, 2)

                if not ((0 < x_overlap < 1) & (0 < y_overlap < 1)):
                    raise RuntimeError("failed to determine overlap percentage from metadata.")

                overlap = (y_overlap, x_overlap)

            width = x_ndim * image.attributes.widthPx - (x_ndim - 1) * round(image.attributes.widthPx * overlap[1])
            height = y_ndim * image.attributes.heightPx - (y_ndim - 1) * round(image.attributes.heightPx * overlap[0])

            image_shape = None
            tiles = np.empty((y_ndim, x_ndim), dtype=object)

            for n in range(axis_sizes['P']):

                tile = image_array[n]
                if tile.ndim > 2:
                    tile = tile[slices]
                if image_shape is None:
                    image_shape = (*tile.shape[:-2], height, width)
                tiles[i[n], j[n]] = tile

            tiling = tiles.shape

        tile_size = (image.attributes.heightPx, image.attributes.widthPx)

    return tiles, tiling, tile_size, overlap, image_shape



import dask.array as da
import numpy as np
from dask import delayed
from deeptile.core import utils


def parse(image, image_shape, tile_size, overlap, slices):

    overlap_size = utils.calculate_overlap_size(np.array(tile_size), np.array(overlap))
    tiling = utils.calculate_tiling(np.array(image_shape), np.array(tile_size), overlap_size)
    tiling = tuple(tiling)
    tile_size = np.ceil(tile_size)
    overlap_size = np.floor(overlap_size)

    tiles = np.empty(tiling, dtype=object)
    gys = []
    gxs = []
    heights = []
    widths = []

    tile_iterator = image.tileIterator(frame=slices,
                                       tile_size=dict(height=tile_size[0], width=tile_size[1]),
                                       tile_overlap=dict(y=overlap_size[0], x=overlap_size[1]))
    lazy_imread = delayed(imread)
    for tile_dict in tile_iterator:
        delayed_reader = lazy_imread(tile_dict)
        shape = (tile_dict['height'], tile_dict['width'])
        tiles[tile_dict['level_y'], tile_dict['level_x']] = da.from_delayed(delayed_reader, shape=shape, dtype=object)
        gys.append(tile_dict['gy'])
        gxs.append(tile_dict['gx'])
        heights.append(tile_dict['height'])
        widths.append(tile_dict['width'])

    gys = gys[::tiling[1]]
    gxs = gxs[:tiling[1]]
    heights = heights[::tiling[1]]
    widths = widths[:tiling[1]]

    v_tile_indices = np.cumsum((gys, heights), axis=0).T.astype(int)
    h_tile_indices = np.cumsum((gxs, widths), axis=0).T.astype(int)
    tile_indices = (v_tile_indices, h_tile_indices)

    v_border_indices = np.mean(v_tile_indices.ravel()[1:-1].reshape(-1, 2), axis=1).astype(int)
    v_border_indices = np.concatenate(([0], v_border_indices, [image_shape[0]]))
    h_border_indices = np.mean(h_tile_indices.ravel()[1:-1].reshape(-1, 2), axis=1).astype(int)
    h_border_indices = np.concatenate(([0], h_border_indices, [image_shape[1]]))
    border_indices = (v_border_indices, h_border_indices)

    return tiles, tiling, tile_indices, border_indices


def imread(tile_dict):

    tile = tile_dict['tile'][:, :, 0]
    tile_dict.release()

    return tile



def test_stitch_image():

    """ Test that DeepTile can stitch an image. """

    import numpy as np
    import tifffile
    from deeptile import load
    from deeptile.extensions import stitch
    from pathlib import Path

    image = str(Path(__file__).parents[1] / 'data' / 'sample.nd2')
    ground_truth = tifffile.imread(str(Path(__file__).parents[1] / 'data' / 'sample_nd2_stitched_sigma_10.tif'))
    ground_truth_no_blend = \
        tifffile.imread(str(Path(__file__).parents[1] / 'data' / 'sample_nd2_stitched_no_blend.tif'))

    dt = load(image)
    tiles = dt.get_tiles()
    stitched = stitch.stitch_image(tiles, blend=True, sigma=10)

    assert stitched.dtype is ground_truth.dtype
    assert np.all(stitched == ground_truth)

    stitched_no_blend = stitch.stitch_image(tiles, blend=False)

    assert stitched_no_blend.dtype is ground_truth_no_blend.dtype
    assert np.all(stitched_no_blend == ground_truth_no_blend)


def test_stitch_masks():

    """ Test that DeepTile can stitch masks. """

    import numpy as np
    import pickle
    import tifffile
    from deeptile import load
    from deeptile.extensions import stitch
    from pathlib import Path

    with open(str(Path(__file__).parents[1] / 'data' / 'sample_tif_masks'), 'rb') as f:
        masks = pickle.load(f)

    ground_truth = \
        tifffile.imread(str(Path(__file__).parents[1] / 'data' / 'sample_tif_stitched_mask_iou_threshold_10.tif'))

    dt = load(np.zeros((1022, 1024)))
    tiles = dt.get_tiles(tile_size=(300, 300), overlap=(0.1, 0.1))
    tiles[:] = masks
    stitched = stitch.stitch_masks(tiles, iou_threshold=0.1)

    assert np.all(stitched == ground_truth)



import numpy as np
from deeptile.core import process, trees
from deeptile.core.data import Tiled
from deeptile.core.iterators import Iterator
from deeptile.core.jobs import Job
from functools import wraps


class Lifted:

    """Lifted class for functions lifted to be applied on Tiled objects."""

    def __new__(cls, func, vectorized, batch_axis, pad_final_batch, batch_size):

        """Lift function.

        Parameters
        ----------
        func : Callable
            Callable for use in tile processing.
        vectorized : bool
            Whether the algorithm is vectorized to support batching.
        batch_axis : bool
            Whether to use the first axis to create batches.
        pad_final_batch : bool
            Whether to pad the final batch to the specified ``batch_size``. If ``func_process`` does not support
            batching, this value is ignored.
        batch_size : int
            Number of tiles in each batch. If ``func`` is not vectorized, this value is ignored.

        Returns
        -------
        lifted_func : Callable
            Lifted function.
        """

        lifted_func = super().__new__(cls)
        lifted_func.__call__ = wraps(func)(lifted_func)
        lifted_func.func = func
        lifted_func.vectorized = vectorized
        lifted_func.batch_axis = batch_axis
        lifted_func.pad_final_batch = pad_final_batch
        lifted_func.batch_size = batch_size

        return lifted_func

    def __call__(self, *args, **kwargs):

        """Apply lifted function on Tiled objects.

        Returns
        -------
        processed_tiles
            Tiles processed by lifted function.
        """

        processed_tiles, variables = self.init(*args, **kwargs)

        n_steps = variables['n_steps']
        for _ in range(n_steps):
            processed_tiles, variables = self.apply(processed_tiles, variables)

        return processed_tiles

    def init(self, *args, **kwargs):

        """Initialize a lifted job.

        Returns
        -------
        processed_tiles
            Tiles processed by lifted function.
        variables : dict
            Dictionary of variables used in the lifted job.
        """

        job_locals = locals()

        arg_indices = [arg_index for arg_index in trees.tree_scan(args)[2]
                       if isinstance(trees.tree_index(args, arg_index), (Iterator, Tiled))]
        kwarg_indices = [kwarg_index for kwarg_index in trees.tree_scan(kwargs)[2]
                         if isinstance(trees.tree_index(kwargs, kwarg_index), (Iterator, Tiled))]
        inputs = [trees.tree_index(args, arg_index) for arg_index in arg_indices] + \
                 [trees.tree_index(kwargs, kwarg_index) for kwarg_index in kwarg_indices]
        tiles = [inp if isinstance(inp, Tiled) else inp.tiles for inp in inputs]
        process.check_compatability(tiles)

        job_locals['args'] = trees.tree_apply(args, arg_indices, lambda ts: Tiled)
        job_locals['kwargs'] = trees.tree_apply(kwargs, kwarg_indices, lambda ts: Tiled)
        job = Job(inputs, 'lifted_func', job_locals)

        reference = tiles[0]
        nonempty_indices = reference.nonempty_indices
        processed_istree = None
        processed_indices = None
        processed_tiles = None

        if self.batch_axis:
            batch_axis_len = reference[nonempty_indices[0][0], nonempty_indices[1][0]].shape[0]
            batch_axis_indices = np.tile(np.arange(batch_axis_len), len(nonempty_indices[0]))
            nonempty_indices = [np.repeat(np.array(indices), batch_axis_len, 0) for indices in nonempty_indices]
            nonempty_indices.append(batch_axis_indices)
            nonempty_indices = tuple(nonempty_indices)
        else:
            batch_axis_len = None

        if self.vectorized:
            n_steps = np.ceil(len(nonempty_indices[0]) / self.batch_size).astype(int)
        else:
            n_steps = len(nonempty_indices[0])

        step = 0

        variables = {
            'args': args,
            'kwargs': kwargs,
            'arg_indices': arg_indices,
            'kwarg_indices': kwarg_indices,
            'job': job,
            'reference': reference,
            'nonempty_indices': nonempty_indices,
            'batch_axis_len': batch_axis_len,
            'processed_istree': processed_istree,
            'processed_indices': processed_indices,
            'n_steps': n_steps,
            'step': step
        }

        return processed_tiles, variables

    def apply(self, processed_tiles, variables):

        """Take one step in a lifted job.

        Parameters
        ----------
        processed_tiles
            Tiles processed by lifted function.
        variables : dict
            Dictionary of variables used in the lifted job.

        Returns
        -------
        processed_tiles
            Tiles processed by lifted function.
        variables : dict
            Dictionary of variables used in the lifted job.
        """

        n_steps = variables['n_steps']
        step = variables['step']

        if step < n_steps:

            args = variables['args']
            kwargs = variables['kwargs']
            arg_indices = variables['arg_indices']
            kwarg_indices = variables['kwarg_indices']
            job = variables['job']
            reference = variables['reference']
            nonempty_indices = variables['nonempty_indices']
            processed_istree = variables['processed_istree']
            processed_indices = variables['processed_indices']

            if self.vectorized:

                batch_axis_len = variables['batch_axis_len']
                
                batch_offset = step * self.batch_size
                batch_indices = tuple(i[batch_offset:batch_offset + self.batch_size] for i in nonempty_indices)

                processed_istree, processed_indices, processed_tiles = \
                    process.process_vectorized(self.func, self.batch_axis, self.pad_final_batch, self.batch_size,
                                               args, kwargs, arg_indices, kwarg_indices,
                                               job, reference, batch_axis_len, processed_istree, processed_indices,
                                               processed_tiles, batch_indices)

            else:

                index = tuple(i[step] for i in nonempty_indices)
                processed_istree, processed_indices, processed_tiles = \
                    process.process_single(self.func, self.batch_axis,
                                           args, kwargs, arg_indices, kwarg_indices,
                                           job, reference, processed_istree, processed_indices, processed_tiles,
                                           index)

            variables['processed_istree'] = processed_istree
            variables['processed_indices'] = processed_indices
            variables['step'] = step + 1

        return processed_tiles, variables


def lift(func, vectorized=False, batch_axis=False, pad_final_batch=False, batch_size=4):

    """Lift function to be applied on Tiled objects.

    Parameters
    ----------
    func : Callable
        Callable for use in tile processing.
    vectorized : bool, optional, default False
        Whether the algorithm is vectorized to support batching.
    batch_axis : bool, optional, default False
        Whether to use the first axis to create batches.
    pad_final_batch : bool, optional, default False
        Whether to pad the final batch to the specified ``batch_size``. If ``func_process`` does not support batching,
        this value is ignored.
    batch_size : int, optional, default 4
        Number of tiles in each batch. If ``func`` is not vectorized, this value is ignored.

    Returns
    -------
    lifted_func : Callable
        Lifted function.
    """

    lifted_func = Lifted(func, vectorized, batch_axis, pad_final_batch, batch_size)

    return lifted_func



import numpy as np


class Iterator:

    """Base Iterator class that mimics the indexing behavior of numpy.ndarray.

    Parameters
    ----------
    tiles : Tiled
        Array of tiles.
    """

    def __init__(self, tiles):

        self.dt = tiles.dt
        self.profile = tiles.profile
        self.tiles = tiles
        self.mask = tiles.mask
        self.metadata = {
            'isimage': False,
            'stackable': False,
        }
        self.nonempty_mask = tiles.nonempty_mask
        self.nonempty_indices = tiles.nonempty_indices
        self.nonempty_indices_tuples = tiles.nonempty_indices_tuples

    def __getitem__(self, index):

        """Placeholder __getitem__ method.

        Parameters
        ----------
        index : tuple of int
            Array index.

        Raises
        ------
        NotImplementedError
            If no __getitem__ method has been set.
        """

        raise NotImplementedError("no __getitem__ method has been set.")


class IndexIterator(Iterator):

    """Iterator subclass for array indices."""

    def __getitem__(self, index):

        """Get array index.

        Parameters
        ----------
        index : tuple of int
            Array index.

        Returns
        -------
        index : tuple of int
            Array index.
        """

        return index


class TileIndicesIterator(Iterator):

    """Iterator subclass for tile indices."""

    def __getitem__(self, index):

        """Get tile index.

        Parameters
        ----------
        index : tuple of int
            Array index.

        Returns
        -------
        tile_index : numpy.ndarray
            Tile index.
        """

        i, j = index
        tile_indices = self.tiles.tile_indices

        tile_index = np.stack((tile_indices[0][i], tile_indices[1][j]))

        return tile_index


class BorderIndicesIterator(Iterator):

    """Iterator subclass for border indices."""

    def __getitem__(self, index):

        """Get border index.

        Parameters
        ----------
        index : tuple of int
            Array index.

        Returns
        -------
        border_index : numpy.ndarray
            Border index.
        """

        i, j = index
        border_indices = self.tiles.border_indices

        border_index = np.stack((border_indices[0][i:i + 2], border_indices[1][j:j + 2]))

        return border_index


class StitchIndicesIterator(Iterator):

    """Iterator subclass for stitch indices."""

    def __getitem__(self, index):

        """Get stitch index.

        Parameters
        ----------
        index : tuple of int
            Array index.

        Returns
        -------
        stitch_index : tuple of int
            Stitch index.
        """

        tile_indices_iterator = self.tiles.tile_indices_iterator
        border_indices_iterator = self.tiles.border_indices_iterator

        index_image = border_indices_iterator[index]
        index_tile = index_image - tile_indices_iterator[index][:, 0:1]
        stitch_index = (*index_image, *index_tile)

        return stitch_index



import numpy as np
from dask.array import Array
from deeptile.core import trees, utils
from deeptile.core.iterators import IndexIterator, TileIndicesIterator, BorderIndicesIterator, StitchIndicesIterator
from deeptile.core.jobs import Job
from functools import cached_property, partial


class Output:

    """Output class for attaching metadata to function outputs.

    Parameters
    ----------
    output : numpy.ndarray
        Output array.
    **kwargs : dict
        Metadata associated with this function output.

    Returns
    -------
    output : Output
        Output array.
    """

    def __init__(self, output, **metadata):

        self.output = output
        self.metadata = metadata


class Data(np.ndarray):

    """numpy.ndarray subclass for storing DeepTile data."""

    def __new__(cls, data, job):

        """Create new Data object.

        Parameters
        ----------
        data : numpy.ndarray or Data
            Data array.
        job : Job
            Job that generated this data object.

        Returns
        -------
        data : Data
            Data array.
        """

        data = np.asarray(data).view(cls)

        data.dt = job.dt
        data.profile = job.profile
        data.job = job

        if data.dt.link_data:
            data.job.output = data

        return data


class Tiled(Data):

    """numpy.ndarray subclass for storing DeepTile tiled data."""

    def __new__(cls, tiles, job, mask=None, isimage=True, stackable=False, tile_scales=None):

        """Create new Tiled object.

        Parameters
        ----------
        tiles : numpy.ndarray or Tiled
            Array of tiles.
        job : Job
            Job that generated this tiled object.
        mask : numpy.ndarray or None, optional, default None
            Boolean mask. If ``None``, a boolean array with all True values will be used.
        isimage : bool, optional, default True
            Whether each tile is an image.
        stackable : bool, optional, default True
            Whether tiles can be stacked.
        tile_scales : tuple of float or None, optional, default None
            Tile scales relative to profile tile sizes. If ``None``, the values will be inferred.

        Returns
        -------
        tiles : Tiled
            Array of tiles.
        """

        tiles = super().__new__(cls, tiles, job)
        tiles.parent = tiles
        tiles.slices = []
        if mask is None:
            mask = np.ones(tiles.profile.tiling, dtype=bool)
        tiles.mask = mask
        tiles.metadata = {
            'isimage': isimage,
            'stackable': stackable,
            'tile_scales': tile_scales
        }

        return tiles

    def __array_finalize__(self, tiles, **kwargs):

        """Finalize Tiled object.

        Parameters
        ----------
        tiles : numpy.ndarray or Tiled
            Array of tiles.
        """

        if tiles is None:
            return
        self.dt = getattr(tiles, 'dt', None)
        self.profile = getattr(tiles, 'profile', None)
        self.job = getattr(tiles, 'job', None)
        self.parent = getattr(tiles, 'parent', None)
        self.slices = getattr(tiles, 'slices', None)
        self.metadata = getattr(tiles, 'metadata', None)
        self.mask = getattr(tiles, 'mask', None)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):

        """Process tiles using a NumPy universal function.

        Parameters
        ----------
        ufunc : numpy.ufunc
            The ufunc object that was called.
        methods : str
            String indicating how the ``ufunc`` was called, either ``"__call__"`` to indicate it was called directly, or
            one of its methods: ``"reduce"``, ``"accumulate"``, ``"reduceat"``, ``"outer"``, or ``"at"``.
        inputs : tuple
            Tuple of the input arguments to the ``ufunc``.
        **kwargs : dict
            Contains any optional or keyword arguments passed to the function. This includes any ``out`` arguments,
            which are always contained in a tuple.

        Returns
        -------
        processed_tiles : Tiled
            Array of tiles after processing with ``ufunc``.
        """

        from deeptile.core import process

        args = (ufunc, method, *inputs)
        arg_indices = [arg_index for arg_index in trees.tree_scan(args)[2]
                       if isinstance(trees.tree_index(args, arg_index), Tiled)]

        job_locals = {
            'args': trees.tree_apply(args, arg_indices, lambda ts: Tiled),
            'kwargs': kwargs
        }
        job = Job(inputs, 'lifted_array_ufunc', job_locals)

        tiles = [inp for inp in inputs if isinstance(inp, Tiled)]
        process.check_compatability(tiles)

        reference = tiles[0]
        nonempty_indices = reference.nonempty_indices
        processed_istree = None
        processed_indices = None
        processed_tiles = None

        for index in zip(*nonempty_indices):

            processed_istree, processed_indices, processed_tiles = \
                process.process_single(reference[index].__array_ufunc__, False,
                                       args, kwargs, arg_indices, [],
                                       job, reference, processed_istree, processed_indices, processed_tiles,
                                       index)

        return processed_tiles

    def __array_function__(self, func, types, args, kwargs):

        """Process tiles using an arbitrary NumPy function.

        Parameters
        ----------
        func : Callable
            Arbitrary callable exposed by NumPy’s public API, which was called in the form func(*args, **kwargs).
        types : tuple
            types is a collection collections.abc.Collection of unique argument types from the original NumPy function
            call that implement __array_function__.
        args : tuple
            Tuple of arguments directly passed on from the original call.
        kwargs : dict
            Dictionary of keyword arguments directly passed on from the original call.

        Returns
        -------
        processed_tiles : Tiled
            Array of tiles after processing with ``func``.
        """

        from deeptile.core import process

        args = (func, types, args, kwargs)
        arg_indices = [arg_index for arg_index in trees.tree_scan(args)[2]
                       if isinstance(trees.tree_index(args, arg_index), Tiled)]
        inputs = [trees.tree_index(args, arg_index) for arg_index in arg_indices]

        job_locals = {
            'args': trees.tree_apply(args, arg_indices, lambda ts: Tiled),
            'kwargs': {}
        }
        job = Job(inputs, 'lifted_array_function', job_locals)

        tiles = [inp for inp in inputs if isinstance(inp, Tiled)]
        process.check_compatability(tiles)

        reference = tiles[0]
        nonempty_indices = reference.nonempty_indices
        processed_istree = None
        processed_indices = None
        processed_tiles = None

        for index in zip(*nonempty_indices):

            processed_istree, processed_indices, processed_tiles = \
                process.process_single(lambda _func, _types, _args, _kwargs:
                                       reference[index].__array_function__(_func, _types, tuple(_args), _kwargs), False,
                                       args, {}, arg_indices, [],
                                       job, reference, processed_istree, processed_indices, processed_tiles,
                                       index)

        return processed_tiles

    def compute(self, batch_axis=False, pad_final_batch=False, batch_size=None, **kwargs):

        """Compute Dask arrays.

        Parameters
        ----------
        batch_axis : bool, optional, default False
            Whether to use the first axis to create batches.
        pad_final_batch : bool, optional, default False
            Whether to pad the final batch to the specified ``batch_size``.
        batch_size : int or None, optional, default None
            Number of tiles in each batch.

        Returns
        -------
        tiles : Tiled
            Array of tiles.
        """

        from deeptile.core.lift import lift

        if isinstance(self[self.nonempty_indices_tuples[0]], Array):

            tiles = lift(lambda tile: Output(tile.compute(**kwargs), **self.metadata),
                         batch_axis=batch_axis, pad_final_batch=pad_final_batch, batch_size=batch_size)(self)
            tiles.job.type = 'compute_dask'
            tiles.job.locals = kwargs

        else:

            tiles = self

        return tiles

    def persist(self, batch_axis=False, pad_final_batch=False, batch_size=None, **kwargs):

        """Persist Dask arrays into memory.

        Parameters
        ----------
        batch_axis : bool, optional, default False
            Whether to use the first axis to create batches.
        pad_final_batch : bool, optional, default False
            Whether to pad the final batch to the specified ``batch_size``.
        batch_size : int or None, optional, default None
            Number of tiles in each batch.

        Returns
        -------
        tiles : Tiled
            Array of tiles.
        """

        from deeptile.core.lift import lift

        if isinstance(self[self.nonempty_indices_tuples[0]], Array):

            tiles = lift(lambda tile: Output(tile.persist(**kwargs), **self.metadata),
                         batch_axis=batch_axis, pad_final_batch=pad_final_batch, batch_size=batch_size)(self)
            tiles.job.type = 'persist_dask'
            tiles.job.locals = kwargs

        else:

            tiles = self

        return tiles

    def pad(self, **kwargs):

        """Pad tiles to the same size.

        Returns
        -------
        tiles : Tiled
            Array of tiles.

        Raises
        ------
        NotImplementedError
            If padding ``mode`` is not supported.
        """

        mode = kwargs.get('mode', 'constant')

        if mode not in ['constant', 'edge', 'linear_ramp',
                        'maximum', 'mean', 'median', 'minimum',
                        'tile_maximum', 'tile_mean', 'tile_median', 'tile_minimum',
                        'reflect', 'symmetric']:

            raise NotImplementedError('Padding mode is not supported.')

        if self.metadata['isimage'] and not self.metadata['stackable']:

            tiles = np.array(self)

            tile_size = self.tile_size
            tile_indices = self.tile_indices
            tile_padding = (tile_size[0] - (tile_indices[0][-1, 1] - tile_indices[0][-1, 0]),
                            tile_size[1] - (tile_indices[1][-1, 1] - tile_indices[1][-1, 0]))

            if tile_padding[0] > 0:
                if (mode in ['reflect', 'symmetric']) and (tiles.shape[0] > 1):
                    for i, (inner_tile, edge_tile) in enumerate(np.moveaxis(tiles[-2:], 1, 0)):
                        if edge_tile is not None:
                            if inner_tile is not None:
                                extension = self.profile.tile_indices[0][-1, 0] - self.profile.tile_indices[0][-2, 0]
                                extended_tile = np.concatenate((inner_tile[..., :extension, :], edge_tile), axis=-2)
                                padded_tile = utils.array_pad(extended_tile, tile_padding[0], -2, **kwargs)
                                padded_tile = padded_tile[..., -tile_size[0]:, :]
                            else:
                                padded_tile = utils.array_pad(edge_tile, tile_padding[0], -2, **kwargs)
                            tiles[-1, i] = padded_tile
                elif mode in ['tile_maximum', 'tile_mean', 'tile_median', 'tile_minimum']:
                    for i, tile in enumerate(tiles[-1]):
                        if tile is not None:
                            constant = None
                            if mode == 'tile_maximum':
                                constant = np.max(tile)
                            elif mode == 'tile_mean':
                                constant = np.mean(tile)
                            elif mode == 'tile_median':
                                constant = np.median(tile)
                            elif mode == 'tile_minimum':
                                constant = np.min(tile)
                            tiles[-1, i] = \
                                utils.array_pad(tile, tile_padding[0], -2, mode='constant', constant_values=constant)
                else:
                    for i, tile in enumerate(tiles[-1]):
                        if tile is not None:
                            tiles[-1, i] = utils.array_pad(tile, tile_padding[0], -2, **kwargs)

            if tile_padding[1] > 0:
                if (mode in ['reflect', 'symmetric']) and (tiles.shape[1] > 1):
                    for i, (inner_tile, edge_tile) in enumerate(tiles[:, -2:]):
                        if edge_tile is not None:
                            if inner_tile is not None:
                                extension = self.profile.tile_indices[1][-1, 0] - self.profile.tile_indices[1][-2, 0]
                                extended_tile = np.concatenate((inner_tile[..., :extension], edge_tile), axis=-1)
                                padded_tile = utils.array_pad(extended_tile, tile_padding[1], -1, **kwargs)
                                padded_tile = padded_tile[..., -tile_size[1]:]
                            else:
                                padded_tile = utils.array_pad(edge_tile, tile_padding[0], -1, **kwargs)
                            tiles[i, -1] = padded_tile
                elif mode in ['tile_maximum', 'tile_mean', 'tile_median', 'tile_minimum']:
                    for i, tile in enumerate(tiles[:, -1]):
                        if tile is not None:
                            constant = None
                            if mode == 'tile_maximum':
                                constant = np.max(tile)
                            elif mode == 'tile_mean':
                                constant = np.mean(tile)
                            elif mode == 'tile_median':
                                constant = np.median(tile)
                            elif mode == 'tile_minimum':
                                constant = np.min(tile)
                            tiles[i, -1] = \
                                utils.array_pad(tile, tile_padding[1], -1, mode='constant', constant_values=constant)
                else:
                    for i, tile in enumerate(tiles[:, -1]):
                        if tile is not None:
                            tiles[i, -1] = utils.array_pad(tile, tile_padding[1], -1, **kwargs)

            job = Job(self, 'pad_tiles', kwargs)

            metadata = self.metadata.copy()
            metadata['stackable'] = True
            tiles = Tiled(tiles, job, self.mask, **metadata)

        else:

            tiles = self

        return tiles

    def unpad(self):

        """Unpad tiles.

        Returns
        -------
        tiles : Tiled
            Array of tiles.
        """

        if self.metadata['isimage'] and self.metadata['stackable']:

            tiles = np.array(self)

            tile_indices = self.tile_indices

            for i, tile in enumerate(tiles[-1]):
                tiles[-1, i] = tile[..., :tile_indices[0][-1, 1] - tile_indices[0][-1, 0], :]
            for i, tile in enumerate(tiles[:, -1]):
                tiles[i, -1] = tile[..., :tile_indices[1][-1, 1] - tile_indices[1][-1, 0]]

            job = Job(self, 'unpad_tiles', {})

            metadata = self.metadata.copy()
            metadata['stackable'] = False
            tiles = Tiled(tiles, job, self.mask, **metadata)

        else:

            tiles = self

        return tiles

    def import_data(self, data, data_type):

        """Import external data and tile using the same tiling profile.

        Parameters
        ----------
        data
            Data to be imported.
        data_type : str
            Data object type.

        Returns
        -------
        tiles : Tiled
            Array of tiles.

        Raises
        ------
        ValueError
            If ``data_type`` is invalid.
        """

        job_locals = locals()
        job_locals.pop('self')
        job_locals.pop('data')

        from deeptile.core.lift import lift

        if data_type == 'image':
            tiles = lift(partial(utils.tile_image, image=data))(self.tile_indices_iterator)
        elif data_type == 'coords':
            tiles = lift(partial(utils.tile_coords, coords=data))(self.tile_indices_iterator)
        else:
            raise ValueError("invalid data object type.")

        tiles.job.type = 'import_data'
        tiles.job.locals = job_locals
        if self.dt.link_data:
            tiles.job.input = data
        else:
            tiles.job.input = None

        return tiles

    @cached_property
    def tile_size(self):

        """Calculate scaled tile size.

        Returns
        -------
        tile_size : tuple of int
            Scaled tile size.
        """

        profile_tile_size = self.profile.tile_size

        if self.metadata['isimage']:
            if self.metadata['stackable']:
                tile_size = self[self.nonempty_indices_tuples[0]].shape[-2:]
            else:
                tile_size = profile_tile_size
        else:
            tile_scales = self.metadata['tile_scales']
            if tile_scales is None:
                tile_size = profile_tile_size
            else:
                tile_size = (round(profile_tile_size[0] * tile_scales[0]), round(profile_tile_size[1] * tile_scales[1]))

        return tile_size

    @cached_property
    def tile_scales(self):

        """Calculate tile scales relative to profile tile sizes.

        Returns
        -------
        image_scales : tuple of float
            Tile scales relative to profile tile sizes.
        """

        profile_tile_size = self.profile.tile_size
        tile_size = self.tile_size
        tile_scales = (tile_size[0] / profile_tile_size[0], tile_size[1] / profile_tile_size[1])

        return tile_scales

    @cached_property
    def image_size(self):

        """Calculate scaled image size.

        Returns
        -------
        image_size : tuple of int
            Scaled image size.
        """

        profile_image_shape = self.dt.image_shape
        tile_scales = self.tile_scales
        image_size = (round(profile_image_shape[-2] * tile_scales[0]), round(profile_image_shape[-1] * tile_scales[1]))

        return image_size

    @cached_property
    def image_scales(self):

        """Calculate image scales relative to profile image size.

        Returns
        -------
        image_scales : tuple of float
            Image scales relative to profile image size.
        """

        profile_image_shape = self.dt.image_shape
        image_size = self.image_size
        image_scales = (image_size[0] / profile_image_shape[-2], image_size[1] / profile_image_shape[-1])

        return image_scales

    @cached_property
    def nonempty_mask(self):

        """Get a mask for nonempty tiles to be processed.

        Returns
        -------
        nonempty_mask : numpy.ndarray
            Mask for nonempty tiles to be processed.

        Raises
        ------
        ValueError
            If there are no nonempty tiles to process.
        """

        nonempty_mask = self.mask * self.profile.nonempty_mask
        if not np.any(nonempty_mask):
            raise ValueError("no nonempty tiles to process.")

        return nonempty_mask

    @cached_property
    def nonempty_indices(self):

        """Get arrays of indices for nonempty tiles to be processed.

        Returns
        -------
        nonempty_indices : tuple of numpy.ndarray
            Arrays of indices for nonempty tiles to be processed.
        """

        nonempty_indices = np.where(self.nonempty_mask)

        return nonempty_indices

    @cached_property
    def nonempty_indices_tuples(self):

        """Get tuples of indices for nonempty tiles to be processed.

        Returns
        -------
        nonempty_indices_tuples : tuple of tuple
            Tuples of indices for nonempty tiles to be processed.
        """

        nonempty_indices_tuples = tuple(zip(*(tuple(indices) for indices in self.nonempty_indices)))

        return nonempty_indices_tuples

    @cached_property
    def tile_indices(self):

        """Calculate scaled tile indices.

        Returns
        -------
        tile_indices : tuple of numpy.ndarray
            Scaled tile indices.
        """

        image_scales = self.image_scales
        profile_tile_indices = self.profile.tile_indices
        tile_indices = (np.rint(profile_tile_indices[0] * image_scales[0]).astype(int),
                        np.rint(profile_tile_indices[1] * image_scales[1]).astype(int))

        return tile_indices

    @cached_property
    def border_indices(self):

        """Calculate scaled border indices.

        Returns
        -------
        border_indices : tuple of numpy.ndarray
            Scaled border indices.
        """

        image_scales = self.image_scales
        profile_border_indices = self.profile.border_indices
        border_indices = (np.rint(profile_border_indices[0] * image_scales[0]).astype(int),
                          np.rint(profile_border_indices[1] * image_scales[1]).astype(int))

        return border_indices

    @cached_property
    def index_iterator(self):

        """Get a Tiled iterator for array indices.

        Returns
        -------
        index_iterator : IndexIterator
            Tiled iterator for array indices.
        """

        index_iterator = IndexIterator(self)

        return index_iterator

    @cached_property
    def tile_indices_iterator(self):

        """Get a Tiled iterator for tile indices.

        Returns
        -------
        tile_indices_iterator : TileIndicesIterator
            Tiled iterator for tile indices.
        """

        tile_indices_iterator = TileIndicesIterator(self)

        return tile_indices_iterator

    @cached_property
    def border_indices_iterator(self):

        """Get a Tiled iterator for border indices.

        Returns
        -------
        border_indices_iterator : BorderIndicesIterator
            Tiled iterator for border indices.
        """

        border_indices_iterator = BorderIndicesIterator(self)

        return border_indices_iterator

    @cached_property
    def stitch_indices_iterator(self):

        """Get a Tiled iterator for stitch indices.

        Returns
        -------
        stitch_indices_iterator : StitchIndicesIterator
            Tiled iterator for stitch indices.
        """

        stitch_indices_iterator = StitchIndicesIterator(self)

        return stitch_indices_iterator

    @cached_property
    def s(self):

        """Get the Slice object for tile-wise slicing.

        Returns
        -------
        s : Slice
            Slice object for tile-wise slicing.
        """

        s = Slice(self)

        return s

    @cached_property
    def m(self):

        """Get the Mask object for masking.

        Returns
        -------
        m : Mask
            Mask object for masking.
        """

        m = Mask(self)

        return m


class Stitched(Data):

    """numpy.ndarray subclass for storing DeepTile stitched data."""

    def __new__(cls, stitched, job):

        """Create new Stitched object.

        Parameters
        ----------
        stitched : numpy.ndarray
            Stitched object.
        job : Job
            Job that generated this stitched object.

        Returns
        -------
        stitched : Stitched
            Stitched object.
        """

        stitched = super().__new__(cls, stitched, job)

        return stitched

    def __array_finalize__(self, stitched, **kwargs):

        """Finalize Stitched object.

        Parameters
        ----------
        stitched : numpy.ndarray
            Stitched object.
        """

        if stitched is None:
            return
        self.dt = getattr(stitched, 'dt', None)
        self.profile = getattr(stitched, 'profile', None)
        self.job = getattr(stitched, 'job', None)


class Slice:

    """Slice class for tile-wise slicing.

    Parameters
    ----------
    tiles : Tiled
        Array of tiles.
    """

    def __init__(self, tiles):

        self.tiles = tiles

    def __getitem__(self, slices):

        """Apply slices to each tile.

        Parameters
        ----------
        slices : tuple
            Tuple of slice objects designating slices to be extracted.

        Returns
        -------
        sliced_tiles : Tiled
            Sliced array of tiles.
        """

        sliced_tiles = self.tiles.copy()
        sliced_tiles[:] = None
        sliced_tiles.slices = self.tiles.slices + [slices]
        nonempty_indices = self.tiles.nonempty_indices_tuples

        for index in nonempty_indices:
            sliced_tiles[index] = self.tiles[index][slices]

        return sliced_tiles


class Mask:

    """Mask class for masking Tiled objects.

    Parameters
    ----------
    tiles : Tiled
        Array of tiles.
    """

    def __init__(self, tiles):

        self.tiles = tiles

    def __getitem__(self, mask):

        """Mask Tiled object.

        Parameters
        ----------
        mask : numpy.ndarray
            Boolean mask.

        Returns
        -------
        masked_tiles : Tiled
            Masked array of tiles.
        """

        mask = np.broadcast_to(mask, self.tiles.shape).astype(bool)

        masked_tiles = self.tiles.copy()
        masked_tiles[:] = None
        masked_tiles[mask] = self.tiles[mask]
        masked_tiles.mask = mask

        return masked_tiles



import dask.array as da
import tifffile
from dask import delayed


def read(image, dask):

    if dask:
        delayed_reader = delayed(tifffile.memmap)(image)
        image = tifffile.memmap(image)
        image = da.from_delayed(delayed_reader, shape=image.shape, dtype=image.dtype)
    else:
        image = tifffile.imread(image)

    return image



import numpy as np
from deeptile.core import trees, utils
from deeptile.core.data import Output, Tiled
from functools import partial


def process_vectorized(func, batch_axis, pad_final_batch, batch_size,
                       args, kwargs, arg_indices, kwarg_indices,
                       job, reference, batch_axis_len, processed_istree, processed_indices, processed_tiles,
                       batch_indices):

    """Process tiles using a vectorized function.

    Parameters
    ----------
    func : Callable
        Callable for use in tile processing.
    batch_axis : bool
        Whether to use the first axis to create batches.
    pad_final_batch : bool
        Whether to pad the final batch to the specified ``batch_size``. If ``func_process`` does not support batching,
        this value is ignored.
    batch_size : int
        Number of tiles in each batch. If ``func`` is not vectorized, this value is ignored.
    args : tuple
        Arguments for ``func``.
    kwargs : dict
        Keyword arguments for ``func``.
    arg_indices : list of tuple
        Leaf indices for ``args``.
    kwarg_indices : list of tuple
        Leaf indices for ``kwargs``.
    job : Job
        Job associated with this processing step.
    reference : Tiled
        Reference array of tiles.
    batch_axis_len : int
        Length of the batch axis.
    processed_istree : bool
        Whether ``processed_tiles`` is a tree.
    processed_indices : list of tuple
        Leaf indices for ``processed_tiles``.
    processed_tiles
        Tiles processed with ``func``.
    batch_indices : tuple of numpy.ndarray
        Batch indices of tiles.

    Returns
    -------
    processed_istree : bool
        Whether ``processed_tiles`` is a tree.
    processed_indices : list of tuple
        Leaf indices for ``processed_tiles``.
    processed_tiles
        Tiles processed with ``func``.
    """

    batch_args, batch_kwargs = get_arguments(pad_final_batch, batch_size,
                                             args, kwargs, arg_indices, kwarg_indices, batch_indices)
    processed_batch_tiles = func(*batch_args, **batch_kwargs)

    processed_istree, processed_indices, processed_tiles = \
        initialize_tree(processed_batch_tiles, job, reference, processed_istree, processed_indices, processed_tiles)

    processed_batch_tiles = trees.tree_apply(processed_batch_tiles, processed_indices, strip_output_wrapper)

    for i_batch, index in enumerate(zip(*batch_indices)):

        processed_tile = trees.tree_apply(processed_batch_tiles, processed_indices, lambda ts: ts[i_batch])

        if processed_istree:
            for processed_index in processed_indices:
                update_tiles(trees.tree_index(processed_tiles, processed_index),
                             trees.tree_index(processed_tile, processed_index),
                             tuple(index[:2]), batch_axis, batch_axis_len)
        else:
            update_tiles(processed_tiles, processed_tile, tuple(index[:2]), batch_axis, batch_axis_len)

    return processed_istree, processed_indices, processed_tiles


def process_single(func, batch_axis,
                   args, kwargs, arg_indices, kwarg_indices,
                   job, reference, processed_istree, processed_indices, processed_tiles,
                   index):

    """Process tiles using a function.

    Parameters
    ----------
    func : Callable
        Callable for use in tile processing.
    batch_axis : bool
        Whether to use the first axis to create batches.
    args : tuple
        Arguments for ``func``.
    kwargs : dict
        Keyword arguments for ``func``.
    arg_indices : list of tuple
        Leaf indices for ``args``.
    kwarg_indices : list of tuple
        Leaf indices for ``kwargs``.
    job : Job
        Job associated with this processing step.
    reference : Tiled
        Reference array of tiles.
    processed_istree : bool
        Whether ``processed_tiles`` is a tree.
    processed_indices : list of tuple
        Leaf indices for ``processed_tiles``.
    processed_tiles
        Tiles processed with ``func``.
    index : tuple of int
        Index of a tile.

    Returns
    -------
    processed_istree : bool
        Whether ``processed_tiles`` is a tree.
    processed_indices : list of tuple
        Leaf indices for ``processed_tiles``.
    processed_tiles
        Tiles processed with ``func``.
    """

    single_args, single_kwargs = get_arguments(False, 1, args, kwargs, arg_indices, kwarg_indices, index)
    processed_tile = func(*single_args, **single_kwargs)

    processed_istree, processed_indices, processed_tiles = \
        initialize_tree(processed_tile, job, reference, processed_istree, processed_indices, processed_tiles)

    processed_tile = trees.tree_apply(processed_tile, processed_indices, strip_output_wrapper)

    if processed_istree:
        for processed_index in processed_indices:
            update_tiles(trees.tree_index(processed_tiles, processed_index),
                         trees.tree_index(processed_tile, processed_index),
                         tuple(index[:2]), batch_axis)
    else:
        update_tiles(processed_tiles, processed_tile, tuple(index[:2]), batch_axis)

    return processed_istree, processed_indices, processed_tiles


def get_arguments(pad_final_batch, batch_size, args, kwargs, arg_indices, kwarg_indices, batch_indices):

    """Replace Tiled arguments with a single tile or batch of tiles.

    Parameters
    ----------
    pad_final_batch : bool
        Whether to pad the final batch to the specified ``batch_size``. If ``func_process`` does not support batching,
        this value is ignored.
    batch_size : int
        Number of tiles in each batch. If ``func`` is not vectorized, this value is ignored.
    args : tuple
        Arguments for ``func``.
    kwargs : dict
        Keyword arguments for ``func``.
    arg_indices : list of tuple
        Leaf indices for ``args``.
    kwarg_indices : list of tuple
        Leaf indices for ``kwargs``.
    batch_indices : tuple of numpy.ndarray
        Batch indices of tiles.

    Returns
    -------
    args : tuple
        Arguments for ``func`` with Tiled objects replaced.
    kwargs : dict
        Keyword arguments for ``func`` with Tiled objects replaced.
    """

    partial_create_batch = partial(create_batch,
                                   pad_final_batch=pad_final_batch, batch_size=batch_size, batch_indices=batch_indices)
    args = trees.tree_apply(args, arg_indices, partial_create_batch)
    kwargs = trees.tree_apply(kwargs, kwarg_indices, partial_create_batch)

    return args, kwargs


def create_batch(tiles, pad_final_batch, batch_size, batch_indices):

    """Create a batch of tiles from a Tiled object.

    Parameters
    ----------
    tiles : Tiled
        Array of tiles.
    pad_final_batch : bool
        Whether to pad the final batch to the specified ``batch_size``. If ``func_process`` does not support batching,
        this value is ignored.
    batch_size : int
        Number of tiles in each batch. If ``func`` is not vectorized, this value is ignored.
    batch_indices : tuple of numpy.ndarray
        Batch indices of tiles.

    Returns
    -------
    batch_tiles
        Batch of tiles.
    """

    if len(batch_indices) == 2:
        batch_tiles = tiles[batch_indices]
    else:
        if isinstance(batch_indices[0], np.ndarray):
            batch_tiles = [tiles[batch_index[:2]][batch_index[2]] for batch_index in zip(*batch_indices)]
            batch_tiles = utils.cast_list_to_array(batch_tiles)
        else:
            batch_tiles = tiles[batch_indices[:2]][batch_indices[2]]

    if isinstance(batch_tiles, Tiled):
        batch_tiles = np.asarray(batch_tiles)
    if isinstance(batch_tiles, np.ndarray) and (batch_tiles.dtype is np.dtype('O')) and tiles.metadata['stackable']:
        batch_tiles = np.stack(batch_tiles)
        if pad_final_batch and (batch_tiles.shape[0] < batch_size):
            batch_tiles = utils.array_pad(batch_tiles, batch_size - batch_tiles.shape[0], 0)

    return batch_tiles


def initialize_tree(processed_tile, job, reference, processed_istree, processed_indices, processed_tiles):

    """Initialize a tree of Tiled objects.

    Parameters
    ----------
    processed_tile
        Processed tile.
    job : Job
        Job associated with this processing step.
    reference : Tiled
        Reference array of tiles.
    processed_istree : bool
        Whether ``processed_tiles`` is a tree.
    processed_indices : list of tuple
        Leaf indices for ``processed_tiles``.
    processed_tiles
        Tiles processed with ``func``.

    Returns
    -------
    processed_istree : bool
        Whether ``processed_tiles`` is a tree.
    processed_indices : list of tuple
        Leaf indices for ``processed_tiles``.
    processed_tiles
        Tiles processed with ``func``.
    """

    if processed_istree is None:
        processed_istree, _, processed_indices = trees.tree_scan(processed_tile)
        processed_tiles = trees.tree_apply(processed_tile, processed_indices,
                                           partial(initialize_tiles, job=job, reference=reference))

    return processed_istree, processed_indices, processed_tiles


def initialize_tiles(processed_tile, job, reference):

    """Initialize a Tiled object.

    Parameters
    ----------
    processed_tile
        Processed tile.
    job : Job
        Job associated with this processing step.
    reference : Tiled
        Reference array of tiles.

    Returns
    -------
    processed_tiles : Tiled
        Array of tiles.
    """

    processed_tiles = np.empty(reference.profile.tiling, dtype=object)

    mask = reference.mask
    stackable = reference.metadata['stackable']

    if isinstance(processed_tile, Output):
        metadata = processed_tile.metadata
        metadata['stackable'] = metadata.get('stackable', stackable)
        processed_tiles = Tiled(processed_tiles, job, mask, **processed_tile.metadata)
    else:
        processed_tiles = Tiled(processed_tiles, job, mask, stackable=stackable)

    return processed_tiles


def strip_output_wrapper(processed_tile):

    """Strip Output wrapper from a processed tile.

    Parameters
    ----------
    processed_tile
        Processed tile.

    Returns
    -------
    processed_tile
        Processed tile stripped of Output wrapper.

    """

    if isinstance(processed_tile, Output):
        processed_tile = processed_tile.output

    return processed_tile


def update_tiles(processed_tiles, processed_tile, index, batch_axis, batch_axis_len=None):

    """Update a Tiled object.

    Parameters
    ----------
    processed_tiles: Tiled
        Array of tiles.
    processed_tile
        Processed tile.
    index : tuple of int
        Index of a tile.
    batch_axis : bool
        Whether to use the first axis to create batches.
    batch_axis_len : int, optional, default None
        Length of the batch axis.
    """

    if batch_axis:

        current_tile = processed_tiles[index]

        if current_tile is None:
                
            processed_tiles[index] = [processed_tile]

        else:

            current_tile.append(processed_tile)

            if len(current_tile) == batch_axis_len:
                
                if processed_tiles.metadata['stackable'] or processed_tiles.metadata['isimage']:
                    processed_tiles[index] = np.stack(current_tile)
                else:
                    processed_tiles[index] = utils.cast_list_to_array(current_tile)

    else:

        processed_tiles[index] = processed_tile


def check_compatability(tiles):

    """Check if the given tiles are compatible.

    Parameters
    ----------
    tiles : list of Tiled
        Array of tiles.

    Raises
    ------
    ValueError
        If no tiles are given.
    ValueError
        If tiles do not all share a common profile.
    ValueError
        If tiles do not all share a common mask.
    """

    if len(tiles) == 0:
        raise ValueError("no tiles are given.")

    profile = None
    mask = None
    for ts in tiles:

        if profile is None:
            profile = ts.profile
        else:
            if ts.profile is not profile:
                raise ValueError(f'tiles must all share a common profile.')

        if mask is None:
            mask = ts.mask
        else:
            if np.any(ts.mask != mask):
                raise ValueError(f'tiles must all share a common mask.')



def test_import_deeptile():

    """ Test that DeepTile can be imported."""

    import deeptile


def test_import_extensions():

    """ Test that extensions can be imported. """

    from deeptile import extensions



def test_load_dask_array():

    """ Test that DeepTile can load a Dask array. """

    import dask.array as da
    import numpy as np
    from deeptile import load

    image = np.random.rand(10, 10)
    image = da.from_array(image)
    dt = load(image, dask=True)

    assert dt.dask is True
    assert dt.image is image

    dt_no_dask = load(image, dask=False)

    assert dt_no_dask.dask is True
    assert dt_no_dask.image is image


def test_load_numpy_array():

    """ Test that DeepTile can load a NumPy array. """

    import dask.array as da
    import numpy as np
    from deeptile import load

    image = np.random.rand(10, 10)
    dt = load(image, dask=True)

    assert dt.dask is True
    assert isinstance(dt.image, da.Array)
    assert np.all(dt.image == image)

    dt_no_dask = load(image, dask=False)

    assert dt_no_dask.dask is False
    assert dt_no_dask.image is image


def test_from_large_image():

    """ Test that DeepTile can load a large_image tile source. """

    import pytest

    try:
        import large_image
    except ImportError:
        pytest.skip('large_image not installed.')

    import dask.array as da
    import large_image
    import numpy as np
    import tifffile
    from deeptile.io import from_large_image
    from pathlib import Path

    image = str(Path(__file__).parents[1] / 'data' / 'sample.tif')
    ground_truth = tifffile.imread(image)

    ts = large_image.getTileSource(image)
    dt = from_large_image(ts)
    tiles = dt.get_tiles(ground_truth.shape[-2:])

    assert dt.dask is None
    assert dt.image is ts
    assert isinstance(tiles[0, 0], da.Array)
    assert np.all(tiles[0, 0] == ground_truth)


def test_load_nd2():

    """ Test that DeepTile can load an ND2 file. """

    import dask.array as da
    import nd2
    import numpy as np
    from deeptile import load
    from pathlib import Path

    image = str(Path(__file__).parents[1] / 'data' / 'sample.nd2')
    ground_truth = nd2.imread(image, dask=True)

    dt = load(image, dask=True)
    tiles = dt.get_tiles()

    assert dt.dask is True
    assert dt.image == image
    assert np.all(tiles[0, 0] == ground_truth[-1]).compute()
    assert np.all(tiles[-1, -1] == ground_truth[0]).compute()
    assert isinstance(tiles[tiles.nonempty_indices_tuples[0]], da.Array)

    dt_no_dask = load(image, dask=False)
    tiles_no_dask = dt_no_dask.get_tiles()

    assert dt_no_dask.dask is False
    assert dt_no_dask.image == image
    assert np.all(tiles_no_dask[0, 0] == ground_truth[-1]).compute()
    assert np.all(tiles_no_dask[-1, -1] == ground_truth[0]).compute()
    assert isinstance(tiles_no_dask[tiles_no_dask.nonempty_indices_tuples[0]], np.ndarray)


def test_load_tiff():

    """ Test that DeepTile can load a TIFF file. """

    import dask.array as da
    import numpy as np
    import tifffile
    from deeptile import load
    from pathlib import Path

    image = str(Path(__file__).parents[1] / 'data' / 'sample.tif')
    ground_truth = tifffile.imread(image)

    dt = load(image, dask=True)

    assert dt.dask is True
    assert dt.image.dtype is ground_truth.dtype
    assert isinstance(dt.image, da.Array)
    assert np.all(dt.image == ground_truth)

    dt_no_dask = load(image, dask=False)

    assert dt_no_dask.dask is False
    assert dt_no_dask.image.dtype is ground_truth.dtype
    assert isinstance(dt_no_dask.image, np.ndarray)
    assert np.all(dt_no_dask.image == ground_truth)



import numpy as np
from deeptile.core.data import Output
from deeptile.core.lift import lift
from deeptile.core.utils import compute_dask
from functools import partial


def cellpose_segmentation(model_parameters, eval_parameters, output_format='masks'):

    """Generate lifted function for the Cellpose segmentation algorithm.

    Parameters
    ----------
    model_parameters : dict
        Dictionary of model parameters.
    eval_parameters : dict
        Dictionary of evaluation parameters.
    output_format : str, optional
        Format of the output. Supported formats are 'masks' and 'polygons'. Default is 'masks'.

    Returns
    -------
    func_segment : Callable
        Lifted function for the Cellpose segmentation algorithm.
    """

    from cellpose.models import CellposeModel
    from cellpose.io import logger_setup
    logger_setup()

    model = CellposeModel(**model_parameters)

    @lift
    def _func_segment(tile, index, tile_index, stitch_index, tiling):

        tile = compute_dask(tile)
        mask = model.eval(tile, **eval_parameters)[0]

        if output_format == 'masks':
            return mask
        elif output_format == 'polygons':
            polygons = mask_to_polygons(mask, index, tile_index, stitch_index, tiling)
            return polygons

    def func_segment(tiles):

        return _func_segment(tiles, tiles.index_iterator, tiles.tile_indices_iterator, tiles.stitch_indices_iterator,
                             tiles.profile.tiling)

    return func_segment


def deepcell_mesmer_segmentation(model_parameters, eval_parameters):

    """Generate lifted function for the DeepCell Mesmer segmentation algorithm.

    Parameters
    ----------
    model_parameters : dict
        Dictionary of model parameters.
    eval_parameters : dict
        Dictionary of evaluation parameters.

    Returns
    -------
    func_segment : Callable
        Lifted function for the DeepCell Mesmer segmentation algorithm.
    """

    from deepcell.applications import Mesmer

    model = Mesmer(**model_parameters)

    @partial(lift, vectorized=True, batch_axis=True)
    def func_segment(tiles):

        tiles = compute_dask(tiles)
        tiles = np.moveaxis(tiles, 1, -1)
        masks = model.predict(tiles, **eval_parameters)
        masks = np.moveaxis(masks, -1, 1)

        return masks

    return func_segment


def deepcell_nuclear_segmentation(model_parameters, eval_parameters):

    """Generate lifted function for the DeepCell Nuclear segmentation algorithm.

    Parameters
    ----------
    model_parameters : dict
        Dictionary of model parameters.
    eval_parameters : dict
        Dictionary of evaluation parameters.

    Returns
    -------
    func_segment : Callable
        Lifted function for the DeepCell Nuclear segmentation algorithm.
    """

    from deepcell.applications import NuclearSegmentation

    model = NuclearSegmentation(**model_parameters)

    @partial(lift, vectorized=True, batch_axis=True)
    def func_segment(tiles):

        tiles = compute_dask(tiles)
        tiles = np.expand_dims(tiles, axis=-1)
        masks = model.predict(tiles, **eval_parameters)[:, :, :, 0]

        return masks

    return func_segment


def deepcell_cytoplasm_segmentation(model_parameters, eval_parameters):

    """Generate lifted function for the DeepCell Cytoplasm segmentation algorithm.

    Parameters
    ----------
    model_parameters : dict
        Dictionary of model parameters.
    eval_parameters : dict
        Dictionary of evaluation parameters.

    Returns
    -------
    func_segment : Callable
        Lifted function for the DeepCell Cytoplasm segmentation algorithm.
    """

    from deepcell.applications import CytoplasmSegmentation

    model = CytoplasmSegmentation(**model_parameters)

    @partial(lift, vectorized=True, batch_axis=True)
    def func_segment(tiles):

        tiles = compute_dask(tiles)
        tiles = np.expand_dims(tiles, axis=-1)
        masks = model.predict(tiles, **eval_parameters)[:, :, :, 0]

        return masks

    return func_segment


def mask_to_polygons(mask, index, tile_index, stitch_index, tiling):

    from rasterio import features

    i_tile, j_tile = tile_index
    _, _, i_stitch, j_stitch = stitch_index

    offset = np.array((j_tile[0], i_tile[0]))

    valid = np.unique(mask[i_stitch[0]:i_stitch[1], j_stitch[0]:j_stitch[1]])
    valid = set(valid[valid > 0])

    raw_polygons = {int(polygon[1]): np.array(polygon[0]['coordinates'][0]).astype(int) + offset for polygon in
                    features.shapes(mask, mask > 0)}

    border_polygons = ({}, {})
    border = set()
    delta = 5
    if index[0] > 0:
        new_border = np.unique(mask[i_stitch[0]:i_stitch[0] + delta, j_stitch[0]:j_stitch[1]])
        new_border = set(new_border[new_border > 0])
        new_border = new_border - border
        border = border | new_border
        border_polygons[0][index[0] - 1] = [raw_polygons[i] for i in new_border]
    if index[0] < tiling[0] - 1:
        new_border = np.unique(mask[i_stitch[1] - delta:i_stitch[1], j_stitch[0]:j_stitch[1]])
        new_border = set(new_border[new_border > 0])
        new_border = new_border - border
        border = border | new_border
        border_polygons[0][index[0]] = [raw_polygons[i] for i in new_border]
    if index[1] > 0:
        new_border = np.unique(mask[i_stitch[0]:i_stitch[1], j_stitch[0]:j_stitch[0] + delta])
        new_border = set(new_border[new_border > 0])
        new_border = new_border - border
        border = border | new_border
        border_polygons[1][index[1] - 1] = [raw_polygons[i] for i in new_border]
    if index[1] < tiling[1] - 1:
        new_border = np.unique(mask[i_stitch[0]:i_stitch[1], j_stitch[1] - delta:j_stitch[1]])
        new_border = set(new_border[new_border > 0])
        new_border = new_border - border
        border = border | new_border
        border_polygons[1][index[1]] = [raw_polygons[i] for i in new_border]

    valid = valid - border
    valid_polygons = [raw_polygons[i] for i in valid]
    polygons = (valid_polygons, border_polygons)
    polygons = Output(polygons, isimage=False, stackable=False)

    return polygons



import numpy as np
from deeptile.core.data import Stitched
from deeptile.core.jobs import Job


def stitch_image(tiles, blend=True, sigma=10):

    """Stitch tiled images into a large image.

    Parameters
    ----------
    tiles : Tiled
        Array of tiled images.
    blend : bool, optional, default True
        Whether to blend tile overlaps.
    sigma : int, optional, default 10
        Sigma bandwidth parameter used to generate sigmoid taper for blending. If ``blend`` is ``False``, this value is
        ignored.

    Returns
    -------
    stitched : Stitched
        Stitched image.
    """

    job_locals = locals()
    job_locals.pop('tiles')

    job = Job(tiles, 'stitch_image', job_locals)

    nonempty_indices = tiles.nonempty_indices_tuples
    first_tile = tiles[nonempty_indices[0]]
    tile_size = first_tile.shape[-2:]
    dtype = first_tile.dtype

    tile_indices_iterator = tiles.tile_indices_iterator
    stitch_shape = (*first_tile.shape[:-2], *tiles.image_size[-2:])

    if blend:

        stitched = np.zeros(stitch_shape, dtype=np.float32)

        avg = np.zeros(stitch_shape[-2:], dtype=np.float32)
        taper = _generate_taper(tile_size, tiles.profile.overlap, sigma)

        for index in nonempty_indices:

            tile_index = tile_indices_iterator[index]

            tile = tiles[index]
            stitch_slice = np.s_[..., tile_index[0, 0]:tile_index[0, 1],
                                 tile_index[1, 0]:tile_index[1, 1]]
            tile_slice = np.s_[..., :tile_index[0, 1] - tile_index[0, 0],
                                    :tile_index[1, 1] - tile_index[1, 0]]
            stitched[stitch_slice] = stitched[stitch_slice] + tile[tile_slice] * taper[tile_slice]
            avg[stitch_slice[1:]] = avg[stitch_slice[1:]] + taper[tile_slice]

        stitched = stitched / (avg + 1e-7)
        stitched = stitched.astype(dtype)

    else:

        stitched = np.zeros(stitch_shape, dtype=dtype)

        stitch_indices_iterator = tiles.stitch_indices_iterator

        for index in nonempty_indices:

            i_image, j_image, i_tile, j_tile = stitch_indices_iterator[index]

            tile = tiles[index]
            tile_crop = tile[..., i_tile[0]:i_tile[1], j_tile[0]:j_tile[1]]
            stitched[..., i_image[0]:i_image[1], j_image[0]:j_image[1]] = tile_crop

    stitched = Stitched(stitched, job)

    return stitched


def stitch_polygons(polygons, iou_threshold=0.1):

    job_locals = locals()
    job_locals.pop('polygons')

    job = Job(polygons, 'stitch_polygons', job_locals)

    valid_polygons = polygons.s[0]
    border_polygons = polygons.s[1]
    tiling = polygons.profile.tiling
    stitched_polygons_list = [polygon for tile in valid_polygons[valid_polygons.nonempty_indices] for polygon in tile]

    for i in range(tiling[0] - 1):
        for j in range(tiling[1]):
            border_u = border_polygons[i, j][0][i]
            border_b = border_polygons[i + 1, j][0][i]

            if j < tiling[1] - 1:
                border_ur = border_polygons[i, j + 1][0][i]
                border_br = border_polygons[i + 1, j + 1][0][i]
                border_u = _overlapping_polygons_rtree(border_ur, border_u, 0.5)
                border_b = _overlapping_polygons_rtree(border_br, border_b, 0.5)

            border_b = _overlapping_polygons_rtree(border_u, border_b, 0.5)
            stitched_polygons_list = stitched_polygons_list + border_u + border_b

    for j in range(tiling[1] - 1):
        for i in range(tiling[0]):
            border_l = border_polygons[i, j][1][j]
            border_r = border_polygons[i, j + 1][1][j]
            border_r = _overlapping_polygons_rtree(border_l, border_r, 0.5)
            stitched_polygons_list = stitched_polygons_list + border_l + border_r

    stitched_polygons = np.empty(len(stitched_polygons_list), dtype=object)
    stitched_polygons[:] = stitched_polygons_list
    stitched_polygons = Stitched(stitched_polygons, job)

    return stitched_polygons


def stitch_masks(masks, iou_threshold=0.1):

    from skimage import measure

    """Stitch tiled masks into a large mask.

    Parameters
    ----------
    masks : Tiled
        Array of tiled masks.
    iou_threshold : float, optional, default 0.1
        IOU score threshold used for mask stitching at tile borders.

    Returns
    -------
    stitched_mask : Stitched
        Stitched mask.
    """

    job_locals = locals()
    job_locals.pop('masks')

    masks = masks.unpad()

    job = Job(masks, 'stitch_masks', job_locals)

    nonempty_indices = masks.nonempty_indices_tuples
    first_mask = masks[nonempty_indices[0]]

    tile_indices = masks.tile_indices
    border_indices = masks.border_indices
    tile_indices_iterator = masks.tile_indices_iterator
    stitch_indices_iterator = masks.stitch_indices_iterator
    mask_shape = (*first_mask.shape[:-2], *masks.image_size[-2:])
    mask_flat_shape = (np.prod(mask_shape[:-2], dtype=int), *mask_shape[-2:])
    stitched_mask = np.zeros(mask_flat_shape, dtype=int)

    for z in range(mask_flat_shape[0]):

        total_count = 0

        for index in nonempty_indices:

            i_image, j_image, i_tile, j_tile = stitch_indices_iterator[index]

            i_clear = i_tile[(0 < i_image) & (i_image < mask_shape[-2])]
            j_clear = j_tile[(0 < j_image) & (j_image < mask_shape[-1])]

            mask = masks[index]
            mask = mask.reshape(-1, *mask.shape[-2:])
            mask = mask[z].copy()
            mask = _clear_border(mask, i_clear, j_clear)

            mask_crop = mask[i_tile[0]:i_tile[1], j_tile[0]:j_tile[1]]
            mask_crop = measure.label(mask_crop)
            count = mask_crop.max()
            mask_crop[mask_crop > 0] += total_count
            total_count += count
            stitched_mask[z, i_image[0]:i_image[1], j_image[0]:j_image[1]] = mask_crop

        border_blobs = _find_border_blobs(masks, tile_indices, border_indices, z)

        for index, blobs in border_blobs.items():

            tile_index = tile_indices_iterator[index]

            mask = masks[index]
            mask = mask.reshape(-1, *mask.shape[-2:])
            mask = mask[z]
            regions = measure.regionprops(mask)
            regions = {region.label: region for region in regions}

            for blob in blobs:

                mask_crop = regions[blob].image
                s = regions[blob].slice
                s_image = (slice(s[0].start + tile_index[0, 0],
                                 s[0].stop + tile_index[0, 0]),
                           slice(s[1].start + tile_index[1, 0],
                                 s[1].stop + tile_index[1, 0]))
                image_crop = stitched_mask[z][s_image]

                if _calculate_iou_score(mask_crop, image_crop > 0) < iou_threshold:
                    image_crop[mask_crop] = total_count + 1
                    total_count += 1

        stitched_mask[z] = measure.label(stitched_mask[z])

    stitched_mask = stitched_mask.reshape(mask_shape)
    stitched_mask = Stitched(stitched_mask, job)

    return stitched_mask


def stitch_coords(coords):

    """Stitch tiled coordinates.

    Parameters
    ----------
    coords : Tiled
        Array of tiled coordinates.

    Returns
    -------
    stitched_coords : Stitched
        Stitched coordinates.
    """

    job = Job(coords, 'stitch_coords', {})

    nonempty_indices = coords.nonempty_indices_tuples
    tile_indices_iterator = coords.tile_indices_iterator
    border_indices_iterator = coords.border_indices_iterator
    first_coord = coords[nonempty_indices[0]]

    if first_coord.dtype is np.dtype('O'):
        batch_axis = True
        n_batches = first_coord.shape[0]
    else:
        batch_axis = False
        n_batches = 1

    stitched_coords = np.empty(n_batches, dtype=object)

    for n in range(n_batches):

        batch_coords = []

        for index in nonempty_indices:

            tile_index = tile_indices_iterator[index]
            border_index = border_indices_iterator[index] - 0.5

            if batch_axis:
                coord = coords[index][n].copy()
            else:
                coord = coords[index].copy()

            coord[:, -2:] = coord[:, -2:] + np.array([tile_index[0, 0], tile_index[1, 0]])
            s = (border_index[0, 0] < coord[:, -2]) & (coord[:, -2] < border_index[0, 1]) & \
                (border_index[1, 0] < coord[:, -1]) & (coord[:, -1] < border_index[1, 1])
            batch_coords.append(coord[s])

        stitched_coords[n] = np.concatenate(batch_coords, axis=0)

    if not batch_axis:
        stitched_coords = stitched_coords[0]
    stitched_coords = Stitched(stitched_coords, job)

    return stitched_coords


def _generate_taper(tile_size, overlap, sigma):

    """(For internal use) Generate taper used for blending tile overlaps.

    Parameters
    ----------
    tile_size : tuple
        Size of each tile.
    overlap : tuple
        Fractions of ``tile_size`` to use for overlap.
    sigma : int
        Sigma bandwidth parameter.

    Returns
    -------
    taper : numpy.array
        Taper used for blending tile overlaps.
    """

    x = np.arange(tile_size[1])
    x = np.abs(x - np.median(x))
    y = np.arange(tile_size[0])
    y = np.abs(y - np.median(y))
    taperx = 1 / (1 + np.exp((x - tile_size[1] * (1 - overlap[1]) / 2) / sigma))
    tapery = 1 / (1 + np.exp((y - tile_size[0] * (1 - overlap[0]) / 2) / sigma))
    taper = tapery[:, None] * taperx

    return taper


def _overlapping_polygons_rtree(polygons1, polygons2, iou_threshold=0.1):

    from rtree import index
    from shapely.geometry import Polygon

    # Initialize an empty R-tree index
    idx = index.Index()

    # Populate R-tree index with each polygon's bounding box in polygons1
    for pos, poly in enumerate(polygons1):
        poly = Polygon(poly)
        idx.insert(pos, poly.bounds)

    # Check for overlapping polygons in polygons2 with polygons1 using R-tree index
    to_remove = set()
    for i in range(len(polygons2)):
        # Iterate over polygons within the bounding box of the current polygon
        poly2 = Polygon(polygons2[i])
        for j in idx.intersection(poly2.bounds):
            # If polygons overlap and IoU > threshold, mark one for removal
            poly1 = Polygon(polygons1[j])
            if poly2.intersects(poly1):
                intersection_area = poly1.intersection(poly2).area
                union_area = poly1.union(poly2).area
                iou = intersection_area / union_area
                if iou > iou_threshold:
                    to_remove.add(i)
                    break  # No need to check further for this polygon in polygons2

    # Create the final lists of polygons after removing the unwanted ones from polygons2
    polygons2 = [polygons2[i] for i in range(len(polygons2)) if i not in to_remove]

    return polygons2


def _calculate_iou_score(a, b):

    """(For internal use) Calculate IOU score.

    Parameters
    ----------
    a : numpy.array
        Boolean array containing object A.
    b : numpy.array
        Boolean array containing object B.

    Returns
    -------
    iou_score : float
        IOU score for objects A and B.
    """

    iou_score = np.sum(a & b) / np.sum(a | b)

    return iou_score


def _find_border_blobs(masks, tile_indices, border_indices, z):

    tile_indices_flat = (tile_indices[0].ravel(), tile_indices[1].ravel())
    border_blobs = dict()

    for axis in range(2):

        if axis == 1:
            masks = masks.T

        for i in range(masks.shape[0]):

            i_image = border_indices[axis][i:i + 2]
            i_tile = i_image - tile_indices[axis][i, 0]

            for j in range(masks.shape[1] - 1):

                j_image = np.flip(tile_indices_flat[1 - axis][2 * j + 1:2 * j + 3])
                offset = tile_indices[1 - axis][j:j + 2, 0]
                j_tile = j_image - offset.reshape(2, 1)
                border_index = border_indices[1 - axis][j + 1] - j_image[0]

                position_l, position_r = None, None

                mask_l_all = masks[i, j]
                if mask_l_all is not None:
                    mask_l_all = mask_l_all.reshape(-1, *mask_l_all.shape[-2:])
                    mask_l_all = mask_l_all[z]
                    if axis == 1:
                        mask_l_all = mask_l_all.T
                        position_l = (j, i)
                    elif axis == 0:
                        position_l = (i, j)
                    border_blobs = _scan_border(border_blobs, mask_l_all, (i_tile, j_tile[0]), position_l, border_index)

                mask_r_all = masks[i, j + 1]
                if mask_r_all is not None:
                    mask_r_all = mask_r_all.reshape(-1, *mask_r_all.shape[-2:])
                    mask_r_all = mask_r_all[z]
                    if axis == 1:
                        mask_r_all = mask_r_all.T
                        position_r = (j + 1, i)
                    elif axis == 0:
                        position_r = (i, j + 1)
                    border_blobs = _scan_border(border_blobs, mask_r_all, (i_tile, j_tile[1]), position_r, border_index)

    return border_blobs


def _scan_border(all_border_blobs, mask_all, position, position_adjacent, border_index):

    if mask_all is not None:

        i, j = position
        mask = mask_all[i[0]:i[1], j[0]:j[1]]
        border = mask[:, border_index]
        blobs = np.unique(border)
        blobs = blobs[blobs > 0]

        if len(blobs) > 0:
            border_blobs = all_border_blobs.get(position_adjacent, set())
            border_blobs.update(blobs)
            all_border_blobs[position_adjacent] = border_blobs

    return all_border_blobs


def _clear_border(mask, i, j):

    for row in i:
        mask = _remove_blob(mask, mask[row])

    for col in j:
        mask = _remove_blob(mask, mask[:, col])

    return mask


def _remove_blob(mask, blobs):

    blobs = np.unique(blobs)
    blobs = blobs[blobs > 0]
    mask[np.isin(mask, blobs)] = 0

    return mask



from deeptile.core import utils
from deeptile.core.data import Tiled
from deeptile.core.jobs import Job
from deeptile.core.profiles import Profile


class DeepTile:

    """Base DeepTile class.

    Parameters
    ----------
    image
        An object containing an image.
    """

    def __init__(self, image):

        self.image = image
        self.image_type = None
        self.image_shape = None
        self.dask = None
        self.link_data = None
        self.profiles = []


class DeepTileArray(DeepTile):

    """DeepTile subclass for arrays.

    Parameters
    ----------
    image : array_like
        An array-like object of an image.
    """

    def __init__(self, image):

        super().__init__(image)
        self.image_type = 'array'

    def get_tiles(self, tile_size, overlap=(0.1, 0.1), slices=(slice(None), )):

        """Split array into tiles.

        Parameters
        ----------
        tile_size : tuple
            Size of each tile.
        overlap : tuple, optional, default (0.1, 0.1)
            Fractions of ``tile_size`` to use for overlap.
        slices : optional, default (slice(None))
            Tuple of slice objects designating slices to be extracted.

        Returns
        -------
        tiles : Tiled
            Array of tiles.
        """

        job_locals = locals()
        job_locals.pop('self')

        image = self.image[slices]
        self.image_shape = image.shape

        tiling, tile_indices, border_indices = utils.calculate_indices(self.image_shape, tile_size, overlap)
        tiles = utils.array_split_2d(image, tile_indices)
        tiles = utils.cast_list_to_array_2d(tiles)
        nonempty_indices = utils.get_nonempty_indices(tiles)

        profile = Profile(self, tiling, tile_size, overlap, slices, nonempty_indices, tile_indices, border_indices)
        job = Job(self.image, 'get_tiles', job_locals, profile)
        tiles = Tiled(tiles, job, stackable=False, tile_scales=(1.0, 1.0))

        return tiles


class DeepTileFunction(DeepTile):

    """DeepTile subclass for functions.

    Parameters
    ----------
    image : Callable
        A function for obtaining image regions.
    image_shape : tuple
        Shape of the image.
    """

    def __init__(self, image, image_shape):

        super().__init__(image)
        self.image_type = 'function'
        self.image_shape = image_shape

    def get_tiles(self, tile_size, overlap=(0.1, 0.1), slices=(slice(None), )):

        """Obtain tiles from function calls.

        Parameters
        ----------
        tile_size : tuple
            Size of each tile.
        overlap : tuple, optional, default (0.1, 0.1)
            Fractions of ``tile_size`` to use for overlap.
        slices : optional, default (slice(None))
            Tuple of slice objects designating slices to be extracted.

        Returns
        -------
        tiles : Tiled
            Array of tiles.
        """

        job_locals = locals()
        job_locals.pop('self')

        from deeptile.sources import function

        tiling, tile_indices, border_indices = utils.calculate_indices(self.image_shape, tile_size, overlap)
        tiles = function.parse(self.image, self.image_shape, tiling, tile_indices, slices)
        nonempty_indices = utils.get_nonempty_indices(tiles)

        profile = Profile(self, tiling, tile_size, overlap, slices, nonempty_indices, tile_indices, border_indices)
        job = Job(self.image, 'get_tiles', job_locals, profile)
        tiles = Tiled(tiles, job, stackable=False, tile_scales=(1.0, 1.0))

        return tiles


class DeepTileLargeImage(DeepTile):

    """DeepTile subclass for large_image tile sources.

    Parameters
    ----------
    image : large_image tile source
        A large_image tile source.
    """

    def __init__(self, image):

        super().__init__(image)
        self.image_type = 'large_image'

    def get_tiles(self, tile_size, overlap=(0.1, 0.1), slices=0):

        """Obtain tiles from large_image tile source.

        Parameters
        ----------
        tile_size : tuple
            Size of each tile.
        overlap : tuple, optional, default (0.1, 0.1)
            Fractions of ``tile_size`` to use for overlap.
        slices : int, optional, default 0
            Frame index designating frame to be extracted.

        Returns
        -------
        tiles : Tiled
            Array of tiles.
        """

        job_locals = locals()
        job_locals.pop('self')

        from deeptile.sources import large_image

        self.image_shape = (self.image.getMetadata()['sizeY'], self.image.getMetadata()['sizeX'])
        tiles, tiling, tile_indices, border_indices = large_image.parse(self.image, self.image_shape,
                                                                        tile_size, overlap, slices)
        nonempty_indices = utils.get_nonempty_indices(tiles)

        profile = Profile(self, tiling, tile_size, overlap, slices, nonempty_indices, tile_indices, border_indices)
        job = Job(self.image, 'get_tiles', job_locals, profile)
        tiles = Tiled(tiles, job, stackable=False, tile_scales=(1.0, 1.0))

        return tiles


class DeepTileND2(DeepTile):

    """DeepTile subclass for ND2 files.

    Parameters
    ----------
    image : str
        An ND2 file.
    """

    def __init__(self, image):

        super().__init__(image)
        self.image_type = 'nd2'
        self.axis_sizes = None
        self.axis_order = None

    def get_tiles(self, overlap=(0.1, 0.1), slices=(slice(None))):

        """Obtain tiles from ND2 file.

        Parameters
        ----------
        overlap : tuple or None, optional, default (0.1, 0.1)
            Fractions to use for overlap. If ``None``, overlap is automatically determined from the ND2 metadata.
        slices : optional, default (slice(None))
            Tuple of slice objects designating slices to be extracted.

        Returns
        -------
        tiles : Tiled
            Array of tiles.
        """

        job_locals = locals()
        job_locals.pop('self')

        from deeptile.sources import nd2

        tiles, tiling, tile_size, overlap, self.image_shape = nd2.parse(self.image, self.dask, self.axis_sizes,
                                                                        self.axis_order, overlap, slices)
        _, tile_indices, border_indices = utils.calculate_indices(self.image_shape, tile_size, overlap)
        nonempty_indices = utils.get_nonempty_indices(tiles)

        profile = Profile(self, tiling, tile_size, overlap, slices, nonempty_indices, tile_indices, border_indices)
        job = Job(self.image, 'get_tiles', job_locals, profile)
        tiles = Tiled(tiles, job, tile_scales=(1.0, 1.0))

        return tiles



from collections.abc import Mapping, Sequence, MutableMapping, MutableSequence


def tree_scan(tree):

    """Scan a tree for all branches and leaves.

    Parameters
    ----------
    tree
        Tree object.

    Returns
    -------
    istree : bool
        Whether the object is a tree.
    branch_indices : list of tuple
        List of all branch indices.
    leaf_indices: list of tuple
        List of all leaf indices.
    """

    istree = _check_istree(tree)

    if istree:

        current_index = [0]
        branch_sizes = [len(tree)]
        branch_indices = []
        leaf_indices = []

        while len(current_index) > 0:

            if current_index[-1] < branch_sizes[-1]:
                current_obj = tree_index(tree, current_index)
                if _check_istree(current_obj):
                    branch_sizes.append(len(current_obj))
                    branch_indices.append(tuple(current_index))
                    current_index.append(0)
                else:
                    leaf_indices.append(tuple(current_index))
                    current_index[-1] += 1
            else:
                current_index.pop()
                if len(current_index) > 0:
                    current_index[-1] += 1
                branch_sizes.pop()

    else:

        branch_indices = []
        leaf_indices = []

    return istree, branch_indices, leaf_indices


def tree_apply(tree, leaf_indices, func):

    """Apply a function to all given leaves of a tree.

    Parameters
    ----------
    tree
        Tree object.
    leaf_indices : list of tuple
        List of leaf indices.
    func : Callable
        Callable to be applied on tree leaves.

    Returns
    -------
    tree
        Tree object.
    """

    if _check_istree(tree):

        if len(leaf_indices) > 0:

            if isinstance(tree, Sequence):
                tree = list(tree)
            elif isinstance(tree, Mapping):
                tree = dict(tree)

            branch_indices = _get_branches_from_leaves(leaf_indices)

            for branch_index in branch_indices:

                branch = tree_index(tree, branch_index)

                if isinstance(branch, Sequence):
                    tree_replace(tree, branch_index, list(branch))
                elif isinstance(branch, Mapping):
                    tree_replace(tree, branch_index, dict(branch))

            for leaf_index in leaf_indices:
                tree_replace(tree, leaf_index, func(tree_index(tree, leaf_index)))

    else:

        tree = func(tree)

    return tree


def tree_replace(tree, index, new_obj):

    """Replace a branch or leaf in a tree.

    Parameters
    ----------
    tree
        Tree object.
    index : tuple of int
        Branch or leaf index.
    new_obj : obj
        New object for replacement.
    """

    branch_index = index[:-1]
    branch = tree_index(tree, branch_index)

    if isinstance(branch, MutableSequence):
        branch[index[-1]] = new_obj

    if isinstance(branch, MutableMapping):
        key = tuple(branch.keys())[index[-1]]
        branch[key] = new_obj


def tree_index(tree, index):

    """Index a tree.

    Parameters
    ----------
    tree
        Tree object.
    index : tuple of int
        Branch or leaf index.

    Returns
    -------
    obj
        Object at given index.
    """

    obj = tree

    for i in index:

        if isinstance(obj, Sequence):
            obj = obj[i]
        elif isinstance(obj, Mapping):
            obj = tuple(obj.values())[i]
        else:
            raise ValueError('tree branches be sequences or mappings.')

    return obj


def _check_istree(obj):

    """(For internal use) Check if an object is a tree.

    Parameters
    ----------
    obj
        Object to be checked.

    Returns
    -------
    istree : bool
        Whether the object is a tree.
    """

    istree = isinstance(obj, (Mapping, Sequence)) and not isinstance(obj, str)

    return istree


def _get_branches_from_leaves(leaf_indices):

    """(For internal use) Get a list of branch indices from leaf indices.

    Parameters
    ----------
    leaf_indices: list of tuple
        List of leaf indices.

    Returns
    -------
    branch_indices : list of tuple
        List of branch indices.
    """

    branch_indices = set(index[:-1] for index in leaf_indices if len(index) > 1)
    num_branches_i = None
    num_branches_f = None
    while (num_branches_f is None) or (num_branches_i < num_branches_f):
        num_branches_i = len(branch_indices)
        branch_indices = branch_indices | set(branch_index[:-1] for branch_index in branch_indices
                                              if len(branch_index) > 1)
        num_branches_f = len(branch_indices)
    branch_indices = list(branch_indices)
    branch_indices.sort(key=lambda index: len(index))

    return branch_indices



import numpy as np
from functools import cached_property


class Profile:

    """Profile class that stores tiling profiles created by a DeepTile object.

    Parameters
    ----------
    dt : DeepTile
        DeepTile object.
    tiling : tuple of int
        Number of tiles in each dimension.
    tile_size : tuple of int
        Size of each tile.
    overlap : tuple of float
        Fractions of ``tile_size`` to use for overlap.
    slices
        Slices to be extracted.
    nonempty_indices : tuple of tuple
        Indices of nonempty tiles.
    tile_indices : tuple of numpy.ndarray
        Indices of tiles.
    border_indices : tuple of numpy.ndarray
        Indices of borders at the middle of tile overlaps.
    """

    def __init__(self, dt, tiling, tile_size, overlap, slices, nonempty_indices, tile_indices, border_indices):

        self.dt = dt
        self.id = len(dt.profiles)
        self.tiling = tiling
        self.tile_size = tile_size
        self.overlap = overlap
        self.slices = slices
        self.nonempty_indices = nonempty_indices
        self.tile_indices = tile_indices
        self.border_indices = border_indices
        self.jobs = []

        dt.profiles.append(self)

    @cached_property
    def nonempty_mask(self):

        """Get a mask for nonempty tiles.

        Returns
        -------
        nonempty_mask : numpy.ndarray
            Mask for nonempty tiles.
        """

        nonempty_mask = np.zeros(self.tiling, dtype=bool)
        nonempty_mask[tuple(zip(*self.nonempty_indices))] = True

        return nonempty_mask



import numpy as np
from collections.abc import Sequence
from dask.array import Array


def to_tuple(obj):

    if isinstance(obj, str) or (not isinstance(obj, Sequence)):
        obj = (obj, )

    return obj


def compute_dask(obj):

    if isinstance(obj, Array):
        obj = obj.compute()

    return obj


def calculate_sliced_shape(shape, slices):

    sliced_shape = list(shape)

    for i, s in enumerate(slices):

        if isinstance(s, int):
            s = slice(s, s + 1)
        elif s is None:
            s = slice(None)

        start, stop, step = s.indices(sliced_shape[i])
        sliced_shape[i] = len(range(start, stop, step))

    sliced_shape = tuple(sliced_shape)

    return sliced_shape


def calculate_tiling(axis_size, tile_size, overlap_size):

    tiling = (axis_size - overlap_size) / (tile_size - overlap_size)
    tiling = np.ceil(tiling).astype(int)

    return tiling


def calculate_overlap_size(tile_size, overlap):

    overlap_size = tile_size * overlap
    overlap_size = np.rint(overlap_size).astype(int)

    return overlap_size


def calculate_indices_1d(axis_size, tile_size, overlap):

    overlap_size = calculate_overlap_size(tile_size, overlap)
    tiling = calculate_tiling(axis_size, tile_size, overlap_size)

    li = np.arange(tiling) * (tile_size - overlap_size)
    ri = np.append(li[:-1] + tile_size, axis_size)
    tile_indices = np.stack((li, ri), axis=1)

    border_indices = np.arange(1, tiling) * (tile_size - overlap_size) + round(0.5 * overlap_size)
    border_indices = np.hstack([0, border_indices, axis_size])

    return tiling, tile_indices, border_indices


def calculate_indices(image_shape, tile_size, overlap):

    v_tiling, v_tile_indices, v_border_indices = calculate_indices_1d(image_shape[-2], tile_size[0], overlap[0])
    h_tiling, h_tile_indices, h_border_indices = calculate_indices_1d(image_shape[-1], tile_size[1], overlap[1])

    tiling = (v_tiling, h_tiling)
    tile_indices = (v_tile_indices, h_tile_indices)
    border_indices = (v_border_indices, h_border_indices)

    return tiling, tile_indices, border_indices


def get_nonempty_indices(tiles):

    nonempty_indices = []

    for index, tile in np.ndenumerate(tiles):
        if tile is not None:
            nonempty_indices.append(index)

    nonempty_indices = tuple(nonempty_indices)

    return nonempty_indices


def axis_take(ary, axis, index):

    return ary[(slice(None), ) * (axis % ary.ndim) + (index, )]


def axis_slice(ary, axis, start, end, step=1):

    return ary[(slice(None), ) * (axis % ary.ndim) + (slice(start, end, step), )]


def array_split(ary, indices, axis):

    sub_arys = [axis_slice(ary, axis, *i) for i in indices]

    return sub_arys


def array_split_2d(ary, indices):

    sub_arys = array_split(ary, indices[0], -2)
    sub_arys = [array_split(sub_ary, indices[1], -1) for sub_ary in sub_arys]

    return sub_arys


def array_pad(ary, padding, axis=0, **kwargs):

    pad_width = [(0, 0)] * ary.ndim
    pad_width[axis] = (0, padding)

    return np.pad(ary, pad_width, **kwargs)


def cast_list_to_array(lst):

    ary = np.empty(len(lst), dtype=object)

    for i, subary in enumerate(lst):
        ary[i] = subary

    return ary


def cast_list_to_array_2d(lst):

    ary = np.empty((len(lst), len(lst[0])), dtype=object)

    for i, sublst in enumerate(lst):
        for j, subary in enumerate(sublst):
            ary[i, j] = subary

    return ary


def tile_image(tile, image):

    tile_index = tile

    image_slice = np.s_[..., tile_index[0, 0]:tile_index[0, 1], tile_index[1, 0]:tile_index[1, 1]]
    tiled_image = image[image_slice]

    return tiled_image


def tile_coords(tile, coords):

    tile_index = tile

    if isinstance(coords, Sequence) or (coords.dtype is np.dtype('O')):
        batch_axis = True
        n_batches = len(coords)
    else:
        coords = (coords,)
        batch_axis = False
        n_batches = 1

    tiled_coords = np.empty(n_batches, dtype=object)

    for n in range(n_batches):

        coord = coords[n]

        s = (tile_index[0, 0] - 0.5 < coord[:, 0]) & (coord[:, 0] < tile_index[0, 1] - 0.5) & \
            (tile_index[1, 0] - 0.5 < coord[:, 1]) & (coord[:, 1] < tile_index[1, 1] - 0.5)
        tiled_coords[n] = coord[s] - tile_index[:, 0]

    if not batch_axis:
        tiled_coords = tiled_coords[0]

    return tiled_coords



class Job:

    """Job class that stores jobs created by a DeepTile object.

    Parameters
    ----------
    job_input
        Input data for the job.
    job_type : str
        Type of job.
    job_locals : dict
        Job local arguments.
    profile : Profile
        New tiling profile used when ``job_input`` is a tile source.
    """

    def __init__(self, job_input, job_type, job_locals, profile=None):

        if profile is None:
            self.profile = job_input[0].profile
        else:
            self.profile = profile
        self.profile.jobs.append(self)

        self.dt = self.profile.dt

        if self.dt.link_data:
            self.input = job_input
        else:
            self.input = None

        self.id = len(self.profile.jobs)
        self.type = job_type
        self.locals = job_locals
        self.output = None



from dask.array import Array
from deeptile import deeptile
from numpy import ndarray
from pathlib import Path


def load(image, dask=True, link_data=True):

    """Load image into a DeepTile object.

    Parameters
    ----------
    image
        An object or path of an image.
    dask : bool, optional, default True
        Whether to use Dask for lazy loading and processing.
    link_data : bool, optional, default True
        Whether to link input and output data to the profile and job objects. Set to ``False`` to reduce memory
        usage.

    Returns
    -------
    dt : DeepTile
        DeepTile object.

    Raises
    ------
    ValueError
        If ``image`` has an unsupported file type.
    ValueError
        If ``image`` is invalid.
    """

    if isinstance(image, Array):
        dask = True
        dt = from_array(image, dask)
    elif isinstance(image, ndarray):
        dt = from_array(image, dask)
    elif Path(image).is_file():
        if image.endswith(('.tif', '.tiff')):
            dt = from_tiff(image, dask)
        elif image.endswith('.nd2'):
            dt = from_nd2(image)
        else:
            raise ValueError('unsupported file type.')
    else:
        raise ValueError("invalid image.")

    dt.dask = dask
    dt.link_data = link_data

    return dt


def from_array(image, dask):

    """Create a DeepTileArray object from an array.

    Parameters
    ----------
    image : array_like
        An array-like object of an image.
    dask : bool, optional, default True
        Whether to use Dask for lazy loading and processing.

    Returns
    -------
    dt : DeepTileArray
        DeepTileArray object.
    """

    from deeptile.sources import array
    image = array.read(image, dask)
    dt = deeptile.DeepTileArray(image)

    return dt


def from_large_image(image):

    """Create a DeepTileLargeImage object from a large_image tile source.

    Parameters
    ----------
    image : large_image tile source
        A large_image tile source.

    Returns
    -------
    dt : DeepTileLargeImage
        DeepTileLargeImage object.
    """

    dt = deeptile.DeepTileLargeImage(image)

    return dt


def from_nd2(image):

    """Create a DeepTileND2 object from an ND2 file.

    Parameters
    ----------
    image : str
        Path to an ND2 file.

    Returns
    -------
    dt : DeepTileND2
        DeepTileND2 object.
    """

    from deeptile.sources import nd2
    image_sizes, axes_order = nd2.read(image)
    dt = deeptile.DeepTileND2(image)
    dt.axis_sizes = image_sizes
    dt.axis_order = axes_order

    return dt


def from_tiff(image, dask):

    """Create a DeepTileArray object from a TIFF file.

    Parameters
    ----------
    image : str
        Path to a TIFF file.
    dask : bool, optional, default True
        Whether to use Dask for lazy loading and processing.

    Returns
    -------
    dt : DeepTileArray
        DeepTileArray object.
    """

    from deeptile.sources import tiff
    image = tiff.read(image, dask)
    dt = deeptile.DeepTileArray(image)

    return dt



import dask.array as da
import numpy as np
from dask import delayed
from deeptile.core import utils


def parse(image, image_shape, tiling, tile_indices, slices):

    tiles = np.empty(tiling, dtype=object)
    sliced_image_shape = utils.calculate_sliced_shape(image_shape, slices)

    lazy_imread = delayed(imread)
    for i in range(tiling[0]):
        for j in range(tiling[1]):
            tile_index = np.stack((tile_indices[0][i], tile_indices[1][j]))
            tile_size = (tile_index[0, 1] - tile_index[0, 0], tile_index[1, 1] - tile_index[1, 0])
            tile_shape = (*sliced_image_shape[:-2], *tile_size)
            delayed_reader = lazy_imread(image, tile_index, slices)
            tiles[i, j] = da.from_delayed(delayed_reader, shape=tile_shape, dtype=object)

    return tiles


def imread(image, tile_index, slices):

    tile = image(tile_index, slices)

    return tile
